@@ -22,7 +22,7 @@
 
 __name__ = 'pylast'
 __version__ = '0.3'
-__revision__ = "$Revision: 194 $"
+__revision__ = "$Revision: 199 $"
 __doc__ = 'A Python interface to Last.fm'
 __author__ = 'Amr Hassan'
 __copyright__ = "Copyright (C) 2008-2009  Amr Hassan"
@@ -39,13 +39,13 @@ SUBMISSION_SERVER = "http://post.audioscrobbler.com:80/"
 
 __proxy = None
 __proxy_enabled = False
-__cache_shelf = None
-__cache_filename = None
+__cache_backend = None
 __last_call_time = 0
 
 import hashlib
 import httplib
 import urllib
+import codecs
 import threading
 from xml.dom import minidom
 import os
@@ -53,6 +53,12 @@ import time
 from logging import info, warn, debug
 import shelve
 import tempfile
+import sys
+
+try:
+	import sqlite3
+except ImportError:
+	pass
 
 STATUS_INVALID_SERVICE = 2
 STATUS_INVALID_METHOD = 3
@@ -109,6 +115,40 @@ SCROBBLE_MODE_PLAYED = "L"
 SCROBBLE_MODE_BANNED = "B"
 SCROBBLE_MODE_SKIPPED = "S"
 
+
+class _ShelfCacheBackend(object):
+	"""Used as a backend for caching cacheable requests."""
+	def __init__(self, file_path = None):
+		self.shelf = shelve.open(file_path)
+	
+	def get_xml(self, key):
+		return self.shelf[key]
+	
+	def set_xml(self, key, xml_string):
+		self.shelf[key] = xml_string
+	
+	def has_key(self, key):
+		return key in self.shelf.keys()
+
+class _SqliteCacheBackend(object):
+	"""Used as a backend for caching cacheable requests."""
+	def __init__(self, file_path = None):
+		self.connection = sqlite3.connect(file_path)
+		self.connection.execute("CREATE TABLE IF NOT EXISTS cache(key TEXT, xml TEXT)")
+		self.connection.commit()
+
+	def get_xml(self, key):
+		row = self.connection.execute("SELECT xml FROM cache WHERE key=?", (key,)).fetchone()
+		return row[0]
+	
+	def set_xml(self, key, xml_string):
+		self.connection.execute("INSERT INTO cache VALUES(?, ?)", (key, xml_string))
+		self.connection.commit()
+
+	def has_key(self, key):
+		row = self.connection.execute("SELECT COUNT(*) FROM cache WHERE key=?", (key,)).fetchone()
+		return row[0] > 0
+		
 class _ThreadedCall(threading.Thread):
 	"""Facilitates calling a function on another thread."""
 	
@@ -154,7 +194,7 @@ class _Request(object):
 		self.params["method"] = method_name
 		
 		if is_caching_enabled():
-			self.shelf = get_cache_shelf()
+			self.cache = _get_cache_backend()
 		
 		if session_key:
 			self.params["sk"] = session_key
@@ -202,14 +242,14 @@ class _Request(object):
 		
 		if not self._is_cached():
 			response = self._download_response()
-			self.shelf[self._get_cache_key()] = response
+			self.cache.set_xml(self._get_cache_key(), response)
 		
-		return self.shelf[self._get_cache_key()]
+		return self.cache.get_xml(self._get_cache_key())
 	
 	def _is_cached(self):
 		"""Returns True if the request is already in cache."""
 		
-		return self.shelf.has_key(self._get_cache_key())
+		return self.cache.has_key(self._get_cache_key())
 		
 	def _download_response(self):
 		"""Returns a response body string from the server."""
@@ -236,9 +276,10 @@ class _Request(object):
 			conn = httplib.HTTPConnection(host=self.HOST_NAME)
 			conn.request(method='POST', url=self.HOST_SUBDIR, body=data, headers=headers)
 		
-		response = conn.getresponse().read()
-		self._check_response_for_errors(response)
-		return response
+		response = conn.getresponse()
+		response_text = unicode(response.read(), "utf-8")
+		self._check_response_for_errors(response_text)
+		return response_text
 		
 	def execute(self, cacheable = False):
 		"""Returns the XML DOM response of the POST Request from the server"""
@@ -248,12 +289,12 @@ class _Request(object):
 		else:
 			response = self._download_response()
 		
-		return minidom.parseString(response)
+		return minidom.parseString(response.encode("utf-8"))
 	
 	def _check_response_for_errors(self, response):
 		"""Checks the response for errors and raises one if any exists."""
 		
-		doc = minidom.parseString(response)
+		doc = minidom.parseString(response.encode("utf-8"))
 		e = doc.getElementsByTagName('lfm')[0]
 		
 		if e.getAttribute('status') != "ok":
@@ -345,7 +386,7 @@ class SessionKeyGenerator(object):
 class _BaseObject(object):
 	"""An abstract webservices object."""
 		
-	def __init__(self, api_key, api_secret, session_key):
+	def __init__(self, api_key, api_secret, session_key=""):
 				
 		self.api_key = api_key
 		self.api_secret = api_secret
@@ -487,11 +528,11 @@ class ServiceException(Exception):
 	"""Exception related to the Last.fm web service"""
 	
 	def __init__(self, lastfm_status, details):
-		self._lastfm_status = lastfm_status
-		self._details = details
+		self.lastfm_status = lastfm_status
+		self.details = details
 	
-	def __str__(self):
-		return self._details
+	def __repr__(self):
+		return self.details
 	
 	def get_id(self):
 		"""Returns the exception ID, from one of the following:
@@ -510,7 +551,7 @@ class ServiceException(Exception):
 			STATUS_TOKEN_EXPIRED = 15
 		"""
 		
-		return self._lastfm_status
+		return self.lastfm_status
 
 class TopItem (object):
 	"""A top item in a list that has a weight. Returned from functions like get_top_tracks() and get_top_artists()."""
@@ -600,12 +641,12 @@ class PlayedTrack (object):
 class Album(_BaseObject, _Taggable):
 	"""A Last.fm album."""
 	
-	def __init__(self, artist, title, api_key, api_secret, session_key):
+	def __init__(self, artist, title, api_key, api_secret, session_key=""):
 		"""
 		Create an album instance.
 		# Parameters:
-			* artist str|Artist: An artist name or an Artist object.
-			* title str: The album title.
+			* artist: An artist name or an Artist object.
+			* title: The album title.
 		"""
 		
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
@@ -724,7 +765,7 @@ class Album(_BaseObject, _Taggable):
 class Artist(_BaseObject, _Taggable):
 	"""A Last.fm artist."""
 	
-	def __init__(self, name, api_key, api_secret, session_key):
+	def __init__(self, name, api_key, api_secret, session_key=""):
 		"""Create an artist object.
 		# Parameters:
 			* name str: The artist's name.
@@ -942,7 +983,7 @@ class Artist(_BaseObject, _Taggable):
 class Event(_BaseObject):
 	"""A Last.fm event."""
 	
-	def __init__(self, event_id, api_key, api_secret, session_key):
+	def __init__(self, event_id, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.id = unicode(event_id)
@@ -1104,7 +1145,7 @@ class Event(_BaseObject):
 class Country(_BaseObject):
 	"""A country at Last.fm."""
 	
-	def __init__(self, name, api_key, api_secret, session_key):
+	def __init__(self, name, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.name = name
@@ -1189,7 +1230,7 @@ class Country(_BaseObject):
 class Library(_BaseObject):
 	"""A user's Last.fm library."""
 	
-	def __init__(self, user, api_key, api_secret, session_key):
+	def __init__(self, user, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		if isinstance(user, User):
@@ -1379,7 +1420,7 @@ class Library(_BaseObject):
 class Playlist(_BaseObject):
 	"""A Last.fm user playlist."""
 	
-	def __init__(self, user, id, api_key, api_secret, session_key):
+	def __init__(self, user, id, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		if isinstance(user, User):
@@ -1511,7 +1552,7 @@ class Tag(_BaseObject):
 	
 	# TODO: getWeeklyArtistChart (too lazy, i'll wait for when someone requests it)
 	
-	def __init__(self, name, api_key, api_secret, session_key):
+	def __init__(self, name, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.name = name
@@ -1646,7 +1687,7 @@ class Tag(_BaseObject):
 class Track(_BaseObject, _Taggable):
 	"""A Last.fm track."""
 	
-	def __init__(self, artist, title, api_key, api_secret, session_key):
+	def __init__(self, artist, title, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		_Taggable.__init__(self, 'track')
 		
@@ -1875,7 +1916,7 @@ class Track(_BaseObject, _Taggable):
 class Group(_BaseObject):
 	"""A Last.fm group."""
 	
-	def __init__(self, group_name, api_key, api_secret, session_key):
+	def __init__(self, group_name, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.name = group_name
@@ -1987,7 +2028,7 @@ class Group(_BaseObject):
 class XSPF(_BaseObject):
 	"A Last.fm XSPF playlist."""
 	
-	def __init__(self, uri, api_key, api_secret, session_key):
+	def __init__(self, uri, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.uri = uri
@@ -2026,7 +2067,7 @@ class XSPF(_BaseObject):
 class User(_BaseObject):
 	"""A Last.fm user."""
 	
-	def __init__(self, user_name, api_key, api_secret, session_key):
+	def __init__(self, user_name, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.name = user_name
@@ -2437,7 +2478,7 @@ class User(_BaseObject):
 		return Library(self, *self.auth_data)
 
 class AuthenticatedUser(User):
-	def __init__(self, api_key, api_secret, session_key):
+	def __init__(self, api_key, api_secret, session_key=""):
 		User.__init__(self, "", api_key, api_secret, session_key);
 	
 	def _get_params(self):
@@ -2591,7 +2632,7 @@ class AuthenticatedUser(User):
 class _Search(_BaseObject):
 	"""An abstract class. Use one of its derivatives."""
 	
-	def __init__(self, ws_prefix, search_terms, api_key, api_secret, session_key):
+	def __init__(self, ws_prefix, search_terms, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self._ws_prefix = ws_prefix
@@ -2630,7 +2671,7 @@ class _Search(_BaseObject):
 class AlbumSearch(_Search):
 	"""Search for an album by name."""
 	
-	def __init__(self, album_name, api_key, api_secret, session_key):
+	def __init__(self, album_name, api_key, api_secret, session_key=""):
 		
 		_Search.__init__(self, "album", {"album": album_name}, api_key, api_secret, session_key)
 	
@@ -2648,7 +2689,7 @@ class AlbumSearch(_Search):
 class ArtistSearch(_Search):
 	"""Search for an artist by artist name."""
 	
-	def __init__(self, artist_name, api_key, api_secret, session_key):
+	def __init__(self, artist_name, api_key, api_secret, session_key=""):
 		_Search.__init__(self, "artist", {"artist": artist_name}, api_key, api_secret, session_key)
 
 	def get_next_page(self):
@@ -2665,7 +2706,7 @@ class ArtistSearch(_Search):
 class TagSearch(_Search):
 	"""Search for a tag by tag name."""
 	
-	def __init__(self, tag_name, api_key, api_secret, session_key):
+	def __init__(self, tag_name, api_key, api_secret, session_key=""):
 		
 		_Search.__init__(self, "tag", {"tag": tag_name}, api_key, api_secret, session_key)
 		
@@ -2684,7 +2725,7 @@ class TrackSearch(_Search):
 	"""Search for a track by track title. If you don't wanna narrow the results down
 	by specifying the artist name, set it to empty string."""
 	
-	def __init__(self, artist_name, track_title, api_key, api_secret, session_key):
+	def __init__(self, artist_name, track_title, api_key, api_secret, session_key=""):
 		
 		_Search.__init__(self, "track", {"track": track_title, "artist": artist_name}, api_key, api_secret, session_key)
 
@@ -2703,7 +2744,7 @@ class VenueSearch(_Search):
 	"""Search for a venue by its name. If you don't wanna narrow the results down
 	by specifying a country, set it to empty string."""
 	
-	def __init__(self, venue_name, country_name, api_key, api_secret, session_key):
+	def __init__(self, venue_name, country_name, api_key, api_secret, session_key=""):
 		
 		_Search.__init__(self, "venue", {"venue": venue_name, "country": country_name}, api_key, api_secret, session_key)
 
@@ -2723,7 +2764,7 @@ class Venue(_BaseObject):
 	
 	# TODO: waiting for a venue.getInfo web service to use.
 	
-	def __init__(self, id, api_key, api_secret, session_key):
+	def __init__(self, id, api_key, api_secret, session_key=""):
 		_BaseObject.__init__(self, api_key, api_secret, session_key)
 		
 		self.id = _number(id)
@@ -2764,7 +2805,7 @@ class Venue(_BaseObject):
 		
 		return list
 		
-def create_new_playlist(title, description, api_key, api_secret, session_key):
+def create_new_playlist(title, description, api_key, api_secret, session_key=""):
 	"""Creates a playlist for the authenticated user and returns it.
 	* title: The title of the new playlist.
 	* description: The description of the new playlist.
@@ -2781,7 +2822,7 @@ def create_new_playlist(title, description, api_key, api_secret, session_key):
 	
 	return Playlist(user, id, api_key, api_secret, session_key)
 
-def get_authenticated_user(api_key, api_secret, session_key):
+def get_authenticated_user(api_key, api_secret, session_key=""):
 	"""Returns the authenticated user."""
 	
 	return AuthenticatedUser(api_key, api_secret, session_key)
@@ -2836,43 +2877,43 @@ def async_call(sender, call, callback = None, call_args = None, callback_args =
 	thread = _ThreadedCall(sender, call, call_args, callback, callback_args)
 	thread.start()
 
-def enable_caching(cache_filename = None):
-	"""Enables caching request-wide for all cachable calls. Uses a shelve.DbfilenameShelf object.
-	* cache_filename: A filename for the db. Defaults to a temporary filename in the tmp directory.
+def enable_caching(file_path = None):
+	"""Enables caching request-wide for all cachable calls.
+	In choosing the backend used for caching, it will try _SqliteCacheBackend first if
+	the module sqlite3 is present. If not, it will fallback to _ShelfCacheBackend which uses shelve.Shelf objects.
+	
+	* file_path: A file path for the backend storage file. If 
+	None set, a temp file would probably be created, according the backend.
 	"""
 	
-	global __cache_shelf
-	global __cache_filename
+	global __cache_backend
 	
-	if not cache_filename:
-		cache_filename = tempfile.mktemp(prefix="pylast_tmp_")
-		
-	__cache_filename = cache_filename
-	__cache_shelf = shelve.open(__cache_filename)
+	if not file_path:
+		file_path = tempfile.mktemp(prefix="pylast_tmp_")
+	
+	if "sqlite3" in sys.modules.keys():
+		__cache_backend = _SqliteCacheBackend(file_path)
+		debug("Caching to Sqlite3 at " + file_path)
+	else:
+		__cache_backend = _ShelfCacheBackend(file_path)
+		debug("Caching to Shelf at " + file_path)
 		
 def disable_caching():
 	"""Disables all caching features."""
 
-	global __cache_shelf	
-	__cache_shelf = None
+	global __cache_backend
+	__cache_backend = None
 
 def is_caching_enabled():
 	"""Returns True if caching is enabled."""
 	
-	global __cache_shelf
-	return not (__cache_shelf == None)
+	global __cache_backend
+	return not (__cache_backend == None)
 
-def get_cache_filename():
-	"""Returns filename of the cache db in use."""
+def _get_cache_backend():
 	
-	global __cache_filename
-	return __cache_filename
-
-def get_cache_shelf():
-	"""Returns the Shelf object used for caching."""
-	
-	global __cache_shelf
-	return __cache_shelf
+	global __cache_backend
+	return __cache_backend
 	
 def _extract(node, name, index = 0):
 	"""Extracts a value from the xml string"""
@@ -2916,32 +2957,32 @@ def _number(string):
 	else:
 		return int(string)
 
-def search_for_album(album_name, api_key, api_secret, session_key):
+def search_for_album(album_name, api_key, api_secret, session_key=""):
 	"""Searches for an album by its name. Returns a AlbumSearch object.
 	Use get_next_page() to retreive sequences of results."""
 	
 	return AlbumSearch(album_name, api_key, api_secret, session_key)
 
-def search_for_artist(artist_name, api_key, api_secret, session_key):
+def search_for_artist(artist_name, api_key, api_secret, session_key=""):
 	"""Searches of an artist by its name. Returns a ArtistSearch object.
 	Use get_next_page() to retreive sequences of results."""
 	
 	return ArtistSearch(artist_name, api_key, api_secret, session_key)
 
-def search_for_tag(tag_name, api_key, api_secret, session_key):
+def search_for_tag(tag_name, api_key, api_secret, session_key=""):
 	"""Searches of a tag by its name. Returns a TagSearch object.
 	Use get_next_page() to retreive sequences of results."""
 	
 	return TagSearch(tag_name, api_key, api_secret, session_key)
 
-def search_for_track(artist_name, track_name, api_key, api_secret, session_key):
+def search_for_track(artist_name, track_name, api_key, api_secret, session_key=""):
 	"""Searches of a track by its name and its artist. Set artist to an empty string if not available.
 	Returns a TrackSearch object.
 	Use get_next_page() to retreive sequences of results."""
 	
 	return TrackSearch(artist_name, track_name, api_key, api_secret, session_key)
 
-def search_for_venue(venue_name, country_name, api_key, api_secret, session_key):
+def search_for_venue(venue_name, country_name, api_key, api_secret, session_key=""):
 	"""Searches of a venue by its name and its country. Set country_name to an empty string if not available.
 	Returns a VenueSearch object.
 	Use get_next_page() to retreive sequences of results."""
@@ -2957,7 +2998,7 @@ def extract_items(topitems_or_libraryitems):
 	
 	return list
 
-def get_top_tags(api_key, api_secret, session_key):
+def get_top_tags(api_key, api_secret, session_key=""):
 	"""Returns a sequence of the most used Last.fm tags as a sequence of TopItem objects."""
 	
 	doc = _Request("tag.getTopTags", dict(), api_key, api_secret, session_key).execute(True)
@@ -2970,7 +3011,7 @@ def get_top_tags(api_key, api_secret, session_key):
 	
 	return list
 
-def get_track_by_mbid(mbid, api_key, api_secret, session_key):
+def get_track_by_mbid(mbid, api_key, api_secret, session_key=""):
 	"""Looks up a track by its MusicBrainz ID."""
 	
 	params = {"mbid": unicode(mbid)}
@@ -2979,7 +3020,7 @@ def get_track_by_mbid(mbid, api_key, api_secret, session_key):
 	
 	return Track(_extract(doc, "name", 1), _extract(doc, "name"), api_key, api_secret, session_key)
 
-def get_artist_by_mbid(mbid, api_key, api_secret, session_key):
+def get_artist_by_mbid(mbid, api_key, api_secret, session_key=""):
 	"""Loooks up an artist by its MusicBrainz ID."""
 	
 	params = {"mbid": unicode(mbid)}
@@ -2988,7 +3029,7 @@ def get_artist_by_mbid(mbid, api_key, api_secret, session_key):
 	
 	return Artist(_extract(doc, "name"), api_key, api_secret, session_key)
 
-def get_album_by_mbid(mbid, api_key, api_secret, session_key):
+def get_album_by_mbid(mbid, api_key, api_secret, session_key=""):
 	"""Looks up an album by its MusicBrainz ID."""
 	
 	params = {"mbid": unicode(mbid)}
@@ -3011,20 +3052,6 @@ def _delay_call():
 	
 	__last_call_time = now
 
-def clear_cache():
-	"""Clears the cache data and starts fresh."""
-	
-	global __cache_dir
-	
-	
-	if not os.path.exists(__cache_dir):
-		return
-	
-	for file in os.listdir(__cache_dir):
-		os.remove(os.path.join(__cache_dir, file))
-
-
-
 # ------------------------------------------------------------
 
 class ScrobblingException(Exception):
@@ -3098,7 +3125,7 @@ class _ScrobblerRequest(object):
 		elif status_line == "BANNED":
 			raise BannedClient()
 		elif status_line == "BADAUTH":
-			raise BadAuthenticationException()
+			raise BadAuthentication()
 		elif status_line == "BADTIME":
 			raise BadTime()
 		elif status_line == "BADSESSION":
